from web_engine import processor, proc_error

class temp_proc(processor):
    def ingress(self, req):
	req.ez_xml_temp = None
	return req
    
    def outgress(self, req):
	if req.ez_xml_temp:
	    req.ez_xml_temp(req.dict_out, req.output_fo)
	else:
	    raise proc_error()
	return req
    pass


class auto_temp_proc(temp_proc):
    def ingress(self, req):
	req.no_temp = False
	return super(auto_temp_proc, self).ingress(req)
    
    def outgress(self, req):
	if req.no_temp:
	    return req
	
	if not req.ez_xml_temp:
	    from tools import partial_import
	    import string
	    
	    try:
		path = req.env['PATH_INFO']
	    except KeyError:
		path = '/'
		pass
	    
	    pparts = path.split('/')
	    pparts = ['ez_templates'] + [part for part in pparts if part]
	    
	    mod, pparts, i = partial_import(pparts)
	    
	    try:
		req.ez_xml_temp = mod.template
	    except AttributeError:
		raise proc_error(404, 'can not load module "' + string.join(pparts, '.') + '", stop at "' + string.join(pparts[:i], '.') + '"')
	    pass
	return super(auto_temp_proc, self).outgress(req)
    pass


def make_stack():
    from web_engine import mime_proc, status_proc, path_to_mod_proc
    stack = []
    stack.append(auto_temp_proc())
    stack.append(mime_proc())
    stack.append(status_proc())
    stack.append(path_to_mod_proc())
    return stack


def run_cgi():
    from web_engine import web_engine
    from cgi_tools import get_form, get_env
    import sys
    
    form = get_form()
    env = get_env()
    processors = make_stack()
    engine = web_engine(processors)
    input_fo, output_fo, error_fo = sys.stdin, sys.stdout, sys.stderr
    engine.process(form, env, input_fo, output_fo, error_fo)
    pass



def ez_import(mod_name):
    mod = __import__(mod_name)
    parts = mod_name.split('.')[1:]
    for part in parts:
	mod = getattr(mod, part)
	pass
    return mod

def partial_import(name_parts):
    assert len(name_parts) > 0
    
    try:
	mod = __import__(name_parts[0])
    except:
	return None, 0
    
    try:
	i = 1
	while i < len(name_parts):
	    if hasattr(mod, 'rewrite_path'):
		name_parts = name_parts[:i] + mod.rewrite_path(name_parts[i:])
		pass
	    name = name_parts[i]
	    __import__(name, mod.__dict__)
	    mod = getattr(mod, name)
	    i = i + 1
	    pass
    except ImportError:
	pass
    return mod, name_parts, i



import unittest
import rpc


def test(stub_o, verify, testcase):
    from pythk.iters import iter_factory, all_combines, all_enum
    astr = 'ab'
    bstr = 'cd'
    argss = [range(i) for i in range(len(astr) + 3)]
    good_factory = iter_factory(lambda: all_combines(astr))
    bad_factory = iter_factory(lambda: all_combines(bstr))
    for args, good, bad in all_enum(argss, good_factory, bad_factory):
	want_success = verify(args, good, bad)
	kws = dict.fromkeys(good + bad)
	try:
	    stub_o(*args, **kws)
	    success = True
	except TypeError, e:
	    success = False
	    pass
	testcase.assertEqual(want_success, success)
	pass
    pass
    
    
class verify(object):
    def __init__(self, varargs, varkws, defaults):
	super(verify, self).__init__()
	self.varargs = varargs
	self.varkws = varkws
	self.defaults = defaults
	pass
	
    def __call__(self, args, good, bad):
	from itertools import izip
	astr='ab'
	bstr = 'cd'
	
	if not self.varargs and len(args) > len(astr):
	    return False
	    
	args_cnt = dict.fromkeys(astr + bstr, 0)
	args_cnt.setdefault(0)
	    
	for k, v in izip(astr, args):
	    args_cnt[k] += 1
	    pass
	for k in good:
	    args_cnt[k] += 1
	    pass
	for k in bad:
	    args_cnt[k] += 1
	    pass
	defaults = self.defaults
	for k in (defaults or []) and astr[-defaults:]:
	    args_cnt[k] = args_cnt[k] or 1
	    pass
	    
	for k in astr:
	    if args_cnt[k] != 1:
		return False
	    pass
	if not self.varkws:
	    for k in bad:
		if args_cnt[k] == 1:
		    return False
		pass
	    pass
	return True
    pass


class rpc_testcase(unittest.TestCase):
    def testRpcStub1(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b, *args, **kws):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(True, True, 0), self)
	pass
    
    def testRpcStub2(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b=1, *args, **kws):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(True, True, 1), self)
	pass
    
    def testRpcStub3(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b, **kws):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(False, True, 0), self)
	pass
    
    def testRpcStub4(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b=1, **kws):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(False, True, 1), self)
	pass
    
    def testRpcStub5(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b, *args):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(True, False, 0), self)
	pass
    
    def testRpcStub6(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b=1, *args):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(True, False, 1), self)
	pass
    
    def testRpcStub7(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(False, False, 0), self)
	pass
    
    def testRpcStub8(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b=1):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(False, False, 1), self)
	pass
    
    def testRpcStub9(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a=1, b=1):
	    pass
	stub_o = facto.get_rpc_stub(foo, ns)
	test(stub_o, verify(False, False, 2), self)
	pass
    
    def testPackUnpack(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	def foo(a, b, *args, **kws):
	    pass
	
	stub_o = facto.get_rpc_stub(foo, ns)
	pkt = stub_o._pack()
	stub_o_r = rpc.rpc_stub._unpack(pkt)
	self.assertEqual(stub_o.__dict__, stub_o_r.__dict__)
	pass
    
    def test_rpc_dispatcher(self):
	import random
	
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	class foo(object):
	    def aoo(self, a, b=3):
		pass
	    def coo(self, a, b, *args):
		return a * b * (((not args) and 1) or args[0])
	    def doo(self, a, b, **kws):
		pass
	    pass
	
	foo_o = foo()
	stub_o = facto.get_rpc_obj_stub(foo_o, ns)
	pkt = stub_o._pack()
	stub_o_r = rpc.rpc_obj_stub._unpack(pkt)
	self.assertEqual(stub_o.aoo._pack(), stub_o_r.aoo._pack())
	self.assertEqual(stub_o.coo._pack(), stub_o_r.coo._pack())
	self.assertEqual(stub_o.doo._pack(), stub_o_r.doo._pack())
	
	dis = rpc.rpc_dispatcher(ns)
	for i in range(1000):
	    n1 = random.random() * 1000
	    n2 = random.random() * 1000
	    pkt = stub_o_r.coo(n1, n2)
	    self.assertEqual(dis.dispatch(pkt), n1 * n2)
	    pass
	for i in range(1000):
	    n1 = random.random() * 1000
	    n2 = random.random() * 1000
	    n3 = random.random() * 1000
	    pkt = stub_o_r.coo(n1, n2, n3)
	    self.assertEqual(dis.dispatch(pkt), n1 * n2 * n3)
	    pass
	pass
    
    def test_make_stubs_in_struct(self):
	ns = rpc.simple_ns()
	facto = rpc.stub_factory(ns)
	class foo(object):
	    def kk(self, abc):
		pass
	    pass
	def bar(a,b):
	    pass
	foo_o = foo()
	boo_o = foo()
	
	old_data = (bar, {'bar': bar, 'foo': foo_o}, 'test', (boo_o, 1))
	data = facto.make_stubs_in_struct(old_data)
	
	self.assert_(data is not old_data)
	self.assert_(isinstance(data, tuple))
	
	self.assert_(isinstance(data[0], rpc.rpc_stub))
	self.assertEqual(data[0], ns.find(obj=bar)[rpc.rpc_stub])
	
	self.assert_(data[1] is old_data[1])
	self.assertEqual(data[0], data[1]['bar'])
	self.assert_(isinstance(data[1]['foo'], rpc.rpc_obj_stub))
	self.assertEqual(data[1]['foo'], ns.find(obj=foo_o)[rpc.rpc_obj_stub])
	
	self.assertEqual(data[2], 'test')
	
	self.assert_(data[3] is not old_data[3])
	self.assert_(isinstance(data[3], tuple))
	self.assert_(isinstance(data[3][0], rpc.rpc_obj_stub))
	self.assertEqual(data[3][1], 1)
	pass
    pass


def suite():
    test = [
        'testRpcStub1', 'testRpcStub2', 'testRpcStub3', 'testRpcStub4',
        'testRpcStub5', 'testRpcStub6', 'testRpcStub7', 'testRpcStub8',
        'testRpcStub9', 'testPackUnpack', 'test_rpc_dispatcher',
	'test_make_stubs_in_struct'
    ]
    cases = [rpc_testcase(t) for t in test]
    suite = unittest.TestSuite(cases)
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite())
    pass




def path_consumer(*trans, **kws):
    n_trans = len(trans)
    n_args = kws.setdefault('n_args', 0) or n_trans
    if isinstance(n_args, tuple):
	at_least = n_args[0]
	at_most = (len(n_args) == 1) and 65535 or n_args[1]
    elif isinstance(n_args, int):
	at_least = n_args
	at_most = (n_args == 0) and 65535 or n_args
    else:
	raise ValueError()
    
    def decorator(func):
	def _call(req):
	    pparts = req.env['PATH_INFO'].split('/')
	    pparts = [p for p in pparts if p]
	    if (len(pparts) < at_least) or (len(pparts) > at_most):
		req.bad_request()
		return
	    
	    try:
		for i, p in enumerate(pparts):
		    if i < n_trans:
			pparts[i] = trans[i](p)
			pass
		    pass
		pass
	    except:
		req.bad_request()
		pass
	    func(*([req] + pparts))
	return _call
    
    return decorator


class uauth:
    @staticmethod
    def may_auth(func):
	from ez_user_auth import uauth
	return lambda req, *args: func(*((req, uauth.authen(req)) + args))
    
    @staticmethod
    def need_auth(func):
	from ez_user_auth import uauth
	def call(req, *args):
	    uid = uauth.authen(req)
	    if uid is None:
		req.precondition_failed()
		return req
	    return func(*((req, uid) + args))
	return call
    
    @staticmethod
    def authorized(req, account):
	from ez_user_auth import uauth
	return uauth.authorized(req, account)
    
    @staticmethod
    def touch(req):
	from ez_user_auth import uauth
	uauth.touch(req)
	pass
    pass



from web_engine import processor, path_to_mod_proc, web_engine
from ez_xml_p import auto_temp_proc
from mod_python import apache, util
import cgi_tools


status_map = {
201: apache.HTTP_CREATED,
204: apache.HTTP_NO_CONTENT,
302: apache.HTTP_MOVED_TEMPORARILY,
400: apache.HTTP_BAD_REQUEST,
401: apache.HTTP_UNAUTHORIZED,
403: apache.HTTP_FORBIDDEN,
404: apache.HTTP_NOT_FOUND,
405: apache.HTTP_METHOD_NOT_ALLOWED,
412: apache.HTTP_PRECONDITION_FAILED
}


class status_proc(processor):
    def outgress(self, req):
	if hasattr(req, 'status'):
	    req.env['_MOD_PYTHON_REQ'].status = status_map[req.status]
	    if (req.status in (201, 302)) and hasattr(req, 'location'):
		req.env['_MOD_PYTHON_REQ'].headers_out['Location'] = req.location
		pass
	    if req.status == 201:
		req.output_fo.write('OK\n')
		pass
	    if req.status != 204:
		req.output_fo.write(str(req.status) + '\n')
		pass
	    req.no_temp = True
	    pass
	return req
    pass


class header_proc(processor):
    def ingress(self, req):
	req.env['HEADERS_OUT'] = {}
	return req
	pass
    
    def outgress(self, req):
	if req.env['HEADERS_OUT']:
	    headers = req.env['HEADERS_OUT']
	    headers_out = req.env['_MOD_PYTHON_REQ'].headers_out
	    for k, v in headers.iteritems():
		headers_out[k] = v
		pass
	    pass
	return req
    pass


class mime_proc(processor):
    def __init__(self):
	super(mime_proc, self).__init__()
	self.charset = charset
	pass
    
    def outgress(self, req):
	if hasattr(req, 'mime_type'):
	    mime = req.mime_type
	else:
	    if self.charset:
		mime = 'text/html; charset=' + self.charset
	    else:
		mime = 'text/html'
		pass
	    pass
	req.env['_MOD_PYTHON_REQ'].content_type = mime
	return req
    pass


class uauth_touch_proc(processor):
    def outgress(self, req):
	from kit import uauth
	uauth.touch(req)
	return req
    pass


def error_hdr(e, req):
    import traceback, sys, log
    import string
    req.env['_MOD_PYTHON_REQ'].content_type = 'text/plain'
    req.env['_MOD_PYTHON_REQ'].status = status_map[e.code]
    _type, value, tb = sys.exc_info()
    req.output_fo.write(str(e.code) + ' ' + e.msg + '\n\n')
    traceback.print_tb(tb, file=req.output_fo)
    
    log.error(str(e.code) + ' ' + e.msg)
    log.error(string.join(traceback.format_tb(tb)))
    pass


def unknow_error_hdr(req):
    import traceback, sys
    import log, string
    req.env['_MOD_PYTHON_REQ'].content_type = 'text/plain'
    req.env['_MOD_PYTHON_REQ'].status = apache.HTTP_INTERNAL_SERVER_ERROR
    _type, value, tb = sys.exc_info()
    req.output_fo.write(str(req.__dict__))
    req.output_fo.write('\n\n')
    req.output_fo.write(_type.__name__ + ': ' + str(value) + '\n')
    traceback.print_tb(tb, file=req.output_fo)
    
    log.error(str(req.__dict__))
    log.error(_type.__name__ + ': ' + str(value))
    log.error(string.join(traceback.format_tb(tb)))
    pass


#
# A work around to support locale specified encoding
# at str()
#
def initialize():
    # this only works for mod_python
    import locale
    global charset
    locale.setlocale(locale.LC_CTYPE, '')
    charset = locale.getlocale(locale.LC_CTYPE)[1]
    # hack str() to support locale specified encoding
    if charset:
	import codecs
	encoder = codecs.getencoder(charset)
	orig_str = str
	class hack_str(orig_str):
	    def __new__(clazz, x=''):
		if isinstance(x, unicode):
		    x = encoder(x)[0]
		elif not isinstance(x, orig_str):
		    x = orig_str.__new__(clazz, x)
		    pass
		return x
	    pass
	orig_isinstance = isinstance
	def hack_isinstance(x, clss):
	    if type(clss) is not tuple:
		return orig_isinstance(x, ((clss is not hack_str) and clss) or orig_str)
	    clss = tuple([((cls is not hack_str) and cls) or orig_str for cls in clss])
	    return orig_isinstance(x, clss)
	
	__builtins__['str'] = hack_str
	__builtins__['isinstance'] = hack_isinstance
	pass
    pass
initialize()

processors = (\
    auto_temp_proc(), \
    mime_proc(), \
    status_proc(), \
    header_proc(), \
    uauth_touch_proc(), \
    path_to_mod_proc() \
    )
engine = web_engine(processors, err_hdr=error_hdr, uerr_hdr=unknow_error_hdr)

def handler(req):
    cgi_form = util.FieldStorage(req)
    form = cgi_tools.get_form_from_FieldStorage(cgi_form)
    env = dict(req.subprocess_env)
    env['_MOD_PYTHON_REQ'] = req
    if req.headers_in.has_key('Cookie'):
	env['HTTP_COOKIE'] = req.headers_in['Cookie']
	pass
    class fake_file(object):
	pass
    fo = fake_file()
    fo.read = req.read
    fo.write = req.write
    engine.process(form, env, fo, fo, fo)
    return apache.OK




import config

if hasattr(config, 'logger'):
    logger = config.logger
    error = logger.error
    warning = logger.warning
    alert = logger.alert
else:
    error = warning = alert = lambda x: None
    pass




def get_form():
    import cgi
    cgi_form = cgi.FieldStorage()
    return get_form_from_FieldStorage(cgi_form)


def get_form_from_FieldStorage(cgi_form):
    form = {}
    for key in cgi_form.keys():
	if not hasattr(cgi_form[key], 'file'):
	    form[key] = cgi_form[key].value
	else:
	    form[key] = cgi_form[key]
	    pass
	pass
    return form


def get_env():
    import os
    os_env = os.environ
    env = dict(os_env)
    return env




'''
A simple implementation of RPC.  It leverages power of Python module 'pickle'
and Dynamic Typing.
'''
from pythk.decorators import catch_on

# ============================================================
# Name service
#
class rpc_ns(object):
    '''
    Name service inteface
    '''
    def regeister(self, obj):
	'''
	Creates a entity and associates with a object.
	'''
	pass
    def unregister(self, en_o):
	pass
    def find(self, obj=None, oid=None):
	'''
	Find a entity with a object or a OID.
	'''
	pass
    pass


class rpc_ns_entity(object):
    '''
    Entity represents a entry create by a name service.
    
    Entity is a entry for a name, it saves attributes for a name.
    A entity is saw as a dictionary that can keep any type of data.
    Each entity is associated with a OID (rpc_ns_entity.oid).  It
    can be used as a key to retrive the entity.
    '''
    def __init__(self, oid, ns=None):
	super(rpc_ns_entity, self).__init__()
	self.oid = oid
	self.ns = ns
	pass
    
    def get_obj(self):
	'''
	Get object of the entity.
	'''
	pass
    
    def __getitem__(self, key):
	pass
    
    def __setitem__(self, key):
	pass
    pass


class simple_ns(rpc_ns):
    '''
    Simple implementation of rpc_ns.
    '''
    class entity(rpc_ns_entity):
	'''
	Simple implementation of rpc_ns_entity.
	'''
	def __init__(self, oid, ns=None):
	    super(simple_ns.entity, self).__init__(oid, ns)
	    self.__attrs = {}
	    pass
	
	def get_obj(self):
	    return self.ns.get_obj(self)
	
	def __getitem__(self, key):
	    return self.__attrs[key]
	
	def __setitem__(self, key, value):
	    self.__attrs[key] = value
	    pass
	pass
    
    def __init__(self):
	super(simple_ns, self).__init__()
	self.__obj_entity = {}
	self.__oid_entity = {}
	pass
    
    @catch_on(KeyError, lambda: None)
    def register(self, obj):
	if self.__obj_entity.has_key(obj):
	    return self.__obj_entity[obj]
	oid = id(obj)
	en_o = simple_ns.entity(oid, self)
	en_o.__obj = obj
	self.__obj_entity[obj] = en_o
	self.__oid_entity[oid] = en_o
	return en_o
    
    def unregister(self, en_o):
	del self.__obj_entity[en_o.get_obj()]
	del self.__oid_entity[en_o.oid]
	pass
    
    @catch_on(KeyError, lambda: None)
    def find(self, obj=None, oid=None):
	if obj:
	    return self.__obj_entity[obj]
	return self.__oid_entity[oid]
    
    @staticmethod
    @catch_on(KeyError, lambda: None)
    def get_obj(en_o):
	return en_o.__obj
    pass


# ============================================================
# RPC stubs
#
class rpc_stub(object):
    '''rpc_stub implements stub for a remote function of method call.'''
    def __init__(self, code, args, varargs, varkws, defaults, im_self=None, end_point=None, piggy=None):
	super(rpc_stub, self).__init__()
	self.code = code
	self.args_names = args
	self.varargs_name = varargs
	self.kws_name = varkws
	self.defaults = defaults
	self.im_self = im_self
	self.end_point = end_point
	self.piggy = piggy
	self.im_name = None
	pass
    
    def __str__(self):
	fmt = '<%s code: %s, args_names: %s, varargs_name: %s, kws_name: %s, defaults: %s, im_self: %s, end_point: %s, piggy: %s, im_name: %s>'
	data = (self.__class__.__name__, self.code, self.args_names, self.varargs_name, self.kws_name, self.defaults, self.im_self, self.end_point, self.piggy, self.im_name)
	data = tuple([str(item) for item in data])
	return fmt % data
    
    def __check_args(self, args, kws):
	from itertools import islice
	
	def error_fmt():
	    if not self.defaults:
		if max_args == 1:
		    fmt = '%s() takes exactly %d argument (%d geven)'
		else:
		    fmt = '%s() takes exactly %d arguments (%d given)'
		    pass
		pass
	    else:
		if max_args == 1:
		    fmt = '%s() takes at most %d argument (%d given)'
		else:
		    fmt = '%s() takes at most %d arguments (%d given)'
		    pass
		pass
	    return fmt
	
	max_args = len(self.args_names)
	n_defaults = (self.defaults or 0) and len(self.defaults)
	
	# calc num of varargs
	n_args = len(args)
	if n_args > max_args:
	    n_varargs = n_args - max_args
	    n_args = max_args
	else:
	    n_varargs = 0
	    pass
	
	if not self.varargs_name and n_varargs:
	    fmt = error_fmt()
	    raise TypeError(fmt % (self.code, max_args, n_args + n_varargs))
	
	if kws:
	    # check multiple values
	    args_names = self.args_names
	    names_it = iter(args_names)
	    for arg_name in islice(names_it, n_args):
		if kws.has_key(arg_name):
		    raise TypeError('%s() got multiple values for keyword argument \'%s\'' % (self.code, arg_name))
		pass
	    
	    # check not specified variables (not in kws)
	    no_default_remains = max_args - n_args - n_defaults
	    if no_default_remains < 0:
		no_default_remains = 0
		pass
	    by_kws = 0
	    for arg_name in islice(names_it, no_default_remains):
		if not kws.has_key(arg_name):
		    fmt = error_fmt()
		    raise TypeError(fmt % (self.code, max_args, n_args + by_kws))
		by_kws += 1
		pass
	    
	    if not self.kws_name:
		# check unknow keywords in kws
		func_args = dict.fromkeys(self.args_names)
		for kw in kws:
		    if not func_args.has_key(kw):
			fmt = '%s() got an unexpected keyword argument \'%s\''
			raise TypeError(fmt % (self.code, kw))
		    pass
		pass
	    pass
	elif (n_args + n_defaults) < max_args:
	    fmt = error_fmt()
	    raise TypeError(fmt % (self.code, max_args, n_args))
	pass
    
    def __call__(self, *args, **kws):
	self.__check_args(args, kws)
	pkt = self._gen_pkt(args, kws)
	return self.send(pkt)
    
    def _gen_pkt(self, args, kws):
	from types import InstanceType, TypeType
	pkt = {'code': self.code, 'args': args, 'kws': kws}
	if self.im_self:
	    pkt['im_self'] = self.im_self
	    pass
	
	return pkt
    
    def send(self, pkt):
	'''
	Send a RPC packet (pkt) to a remote server.
	
	pkt is a dictionary of {'code': ..., 'args': (...), 'kws': {...}}
	'''
	return pkt
    
    @staticmethod
    def _unpack(pkt, stub_clazz=None):
	stub_clazz = stub_clazz or rpc_stub
	
	code = pkt['code']
	args_names = pkt['args_names']
	varargs_name = pkt.setdefault('varargs_name', None)
	kws_name = pkt.setdefault('kws_name', None)
	defaults = pkt.setdefault('defaults', None)
	im_self = pkt.setdefault('im_self', None)
	end_point = pkt.setdefault('end_point', None)
	piggy = pkt.setdefault('piggy')
	im_name = pkt.setdefault('im_name')
	
	stub_o = stub_clazz(code, args_names, varargs_name, kws_name, defaults, im_self, end_point, piggy)
	stub_o.im_name = im_name
	return stub_o
    
    def _pack(self):
	pkt = {'code': self.code, 'args_names': self.args_names}
	if self.varargs_name:
	    pkt['varargs_name'] = self.varargs_name
	    pass
	if self.kws_name:
	    pkt['kws_name'] = self.kws_name
	    pass
	if self.defaults:
	    pkt['defaults'] = self.defaults
	    pass
	if self.im_self:
	    pkt['im_self'] = self.im_self
	    pass
	if self.end_point:
	    pkt['end_point'] = self.end_point
	    pass
	if self.piggy:
	    pkt['piggy'] = self.piggy
	    pass
	if self.im_name:
	    pkt['im_name'] = self.im_name
	return pkt
    pass


class rpc_obj_stub(object):
    '''
    Implements RPC object stub.
    '''
    def __init__(self, obj, method_stubs, end_point=None, piggy=None):
	super(rpc_obj_stub, self).__init__()
	self._obj = obj
	self._end_point = end_point
	self._piggy = piggy
	self._method_stubs = method_stubs
	for method_stub in method_stubs:
	    setattr(self, method_stub.im_name, method_stub)
	    pass
	pass
    
    def __str__(self):
	import string
	fmt = '<%s obj: %s, end_point: %s, piggy: %s, method_stubs: %s>'
	data = (self.__class__.__name__, self._obj, self._end_point, self._piggy, self._method_stubs)
	data = tuple([str(item) for item in data])
	return fmt % data
    
    _stub_clazz = rpc_stub
    
    @staticmethod
    def _unpack(pkt, obj_stub_clazz=None):
	obj_stub_clazz = obj_stub_clazz or rpc_obj_stub
	stub_clazz = obj_stub_clazz._stub_clazz
	method_unpack = stub_clazz._unpack
	
	obj = pkt['obj']
	method_stubs = [method_unpack(m, stub_clazz) for m in pkt['method_stubs']]
	end_point = pkt.setdefault('end_point', None)
	piggy = pkt.setdefault('piggy', None)
	
	return obj_stub_clazz(obj, method_stubs, end_point, piggy)
    
    def _pack(self):
	pkt = {'obj': self._obj, 'method_stubs': [m._pack() for m in self._method_stubs]}
	if self._end_point:
	    pkt['end_point'] = self._end_point
	    pass
	if self._piggy:
	    pkt['piggy'] = self._piggy
	    pass
	return pkt
    pass


# ============================================================
# RPC stub factory & dispatcher
#
class stub_factory(object):
    '''
    Factory provides a set of functions to create stubs.
    
    Stub_factory associates ns and stubs.  It creates stubs and registers
    them with a ns for later using by rpc_dispatcher instances.
    '''
    def __init__(self, ns, stub_clazz=rpc_stub, obj_stub_clazz=rpc_obj_stub, end_point=None):
	super(stub_factory, self).__init__(self)
	self.ns = ns
	self.stub_clazz = stub_clazz
	self.obj_stub_clazz = obj_stub_clazz
	self.end_point = end_point
	pass
    
    def make_stubs_in_struct(self, d):
	'''
	Find user objects of user defined types, functions, and methods
	in a container, if 'd' is a contiainer, and translates them into
	RPC stubs.
	
	Arguments of a RPC are not always native or builtin types.  For
	user defined types, functions, or methods, we need to translate
	it into a rpc_obj_stub.
	
	Note!!  Translation is doing in the place.  Original objects may
	be modified.
	'''
	from types import FunctionType, MethodType
	
	back_tuples = []
	
	class edge(object):
	    def __init__(self, node, parent_edge=None, parent_key=None):
		self.node = node
		self.parent_edge = parent_edge
		self.parent_key = parent_key
		pass
	    
	    def update_node(self, v):
		pe = self.parent_edge
		pnode = pe.node
		if isinstance(pnode, tuple):
		    back_tuples.append(pe)
		    pnode = list(pnode)
		    pe.update_node(pnode)
		    pass
		self.node = pnode[self.parent_key] = v
		pass
	    pass
	
	root = [d]
	root_edge = edge(root)
	
	taskq = [root_edge]
	for edge_o in taskq:
	    node = edge_o.node
	    if isinstance(node, dict):
		new_tasks = [edge(v, edge_o, k) for k, v in node.iteritems()]
		taskq += new_tasks
	    elif isinstance(node, (list, tuple)):
		new_tasks = [edge(v, edge_o, i) for i, v in enumerate(node)]
		taskq += new_tasks
	    elif isinstance(node, (FunctionType, MethodType)):
		edge_o.update_node(self.get_rpc_stub(node))
	    elif not isinstance(node, (float, int, long, str, unicode)):
		edge_o.update_node(self.get_rpc_obj_stub(node))
		pass
	    pass
	
	for edge_o in back_tuples:
	    edge_o.update_node(tuple(edge_o.node))
	    pass
	return root[0]
    
    def get_rpc_stub(self, callable_o, piggy=None):
	'''
	Translate a function or method into a stub.
	'''
	import inspect
	
	ns = self.ns
	
	code_en_o = ns.find(callable_o)
	if code_en_o:
	    return code_en_o[rpc_stub]
	
	if inspect.ismethod(callable_o):
	    for k, v in inspect.getmembers(callable_o):
		if k == 'im_func':
		    func_obj = v
		elif k == 'im_self':
		    im_self = v
		    pass
		pass
	    im_self = im_self and (ns.find(obj=im_self) or ns.register(im_self)).oid
	else:
	    im_self = None
	    func_obj = callable_o
	    pass
	
	args, varargs, varkws, defaults = inspect.getargspec(func_obj)
	if im_self:
	    args = args[1:]
	    pass
	code_en_o = ns.find(obj=callable_o) or ns.register(callable_o)
	code_oid = code_en_o.oid
	stub_o = self.stub_clazz(code_oid, args, varargs, varkws, defaults, im_self, self.end_point, piggy)
	code_en_o[rpc_stub] = stub_o
	return stub_o
    
    def get_rpc_obj_stub(self, obj, piggy=None):
	'''
	Tranlsate a user defined object into a RPC stub.
	
	Methods in the object are translated into stubs.  Other attributes
	can not access by a remote client directly.  All accessing is performed
	by methods through RPC.
	'''
	from inspect import ismethod
	
	ns = self.ns
	
	obj_en_o = ns.find(obj=obj)
	if obj_en_o:
	    return obj_en_o[rpc_obj_stub]
	obj_en_o = ns.register(obj)
	
	method_stubs = []
	for attr in dir(obj):
	    if attr.startswith('_'):
		continue
	    attr_v = getattr(obj, attr)
	    if ismethod(attr_v):
		method_stub = self.get_rpc_stub(attr_v, piggy)
		method_stub.im_name = attr
		method_stubs.append(method_stub)
		pass
	    pass
	obj_oid = obj_en_o.oid
	stub_o = self.obj_stub_clazz(obj_oid, method_stubs, self.end_point, piggy)
	obj_en_o[rpc_obj_stub] = stub_o
	return stub_o
    pass


class rpc_dispatcher(object):
    '''
    Receives a packet for a RPC and calls repective function or method.
    
    A packet is sent by a function or method stub.  When a client starts
    a RPC by calling stub, then rpc_stub._send() is called with a packet.
    rpc_stub._send() sent the packet to the server site.  The server
    site calls a rpc_dispatcher to dispatch to the respective function or
    method.
    '''
    def __init__(self, ns):
	'''
	Need a ns (name server) to map a code OID to the function or
	method object.
	'''
	super(rpc_dispatcher, self).__init__()
	self.ns = ns
	pass
    
    def dispatch(self, pkt):
	'''
	Receives a packet and calls repective function or method.
	'''
	code = pkt['code']
	code_en_o = self.ns.find(oid=code)
	if not code_en_o:
	    raise ValueError('oid 0x%x is invalid' % (code,))
	callo = code_en_o.get_obj()
	args = pkt['args']
	kws = pkt['kws']
	return callo(*args, **kws)
    pass




from mod_python import apache

def error(m):
    apache.log_error(m, apache.APLOG_ERR)
    pass

def warning(m):
    apache.log_error(m, apache.APLOG_WARNING)
    pass

def alert(m):
    apache.log_error(m, apache.APLOG_ALERT)
    pass




class ez_user_auth(object):
    cookie_key = 'EZ_USER_AUTH'
    env_cookie = 'HTTP_COOKIE'
    
    def __init__(self, secret, timeout):
	self.secret = secret
	self.timeout = timeout
	pass
    
    def authorized(self, req, account):
	from Cookie import BaseCookie
	
	cipher = self.__auth_encode(account)
	
	cookie = BaseCookie()
	cookie[self.cookie_key] = cipher
	c = cookie[self.cookie_key]
	c['path'] = '/'
	c['max-age'] = str(self.timeout)
	
	headers = req.env['HEADERS_OUT']
	headers['Set-Cookie'] = cookie.output(header='')
	pass
    
    def authen(self, req):
	from Cookie import BaseCookie
	if not req.env.has_key(self.env_cookie):
	    return None
	cookie = BaseCookie(req.env[self.env_cookie])
	if not cookie.has_key(self.cookie_key):
	    return None
	auth_code = cookie[self.cookie_key].value
	return self.__auth_decode(auth_code)
    
    def touch(self, req):
	if not req.env.has_key(self.env_cookie):
	    return
	
	from Cookie import BaseCookie
	cookie = BaseCookie(req.env[self.env_cookie])
	if not cookie.has_key(self.cookie_key):
	    return
	
	c = cookie[self.cookie_key]
	if self.__auth_decode(c.value):
	    c['max-age'] = str(self.timeout)
	    c['path'] = '/'
	    account = c.value.split(':')[0]
	    cookie[self.cookie_key] = self.__auth_encode(account)
	    
	    headers = req.env['HEADERS_OUT']
	    req.env[self.env_cookie] = headers['Set-Cookie'] = cookie.output(header='')
	    pass
	pass
    
    def __auth_encode(self, account):
	import time, md5
	
	ts = str(int(time.time()))
	key = md5.md5(account + self.secret + ts).hexdigest()
	return account + ':' + key + ':' + ts
    
    def __auth_decode(self, auth_code):
	import md5, time
	pparts = auth_code.split(':')
	account = pparts[0]
	key = pparts[1]
	ts = pparts[2]
	if (time.time() - int(ts)) > self.timeout:
	    return None
	
	check_key = md5.md5(account + self.secret + ts).hexdigest()
	if check_key == key:
	    return account
	return None
    pass


import config
uauth = ez_user_auth(config.auth_secret, config.auth_timeout)




from distutils.core import setup

setup(name='ez_web',
      version='0.1',
      description='A light-weight WEB framework for Python',
      long_description='''
      ez_web is a light-weight WEB framework.  It is desinged
      for modification.  Programmer can start to modify it
      in a short time.''',
      author='Thinker K.F. Li',
      author_email='thinker@branda.to',
      url='https://opensvn.csie.org/traccgi/PumperWeb/wiki/ez_web',
      packages=['ez_web'],
      package_dir={'ez_web': '.'},
      requires=['pythk', 'ez_xml', 'mod_python'],
      classifiers=['License :: OSI Approved :: BSD License',
                   'Programming Language :: Python',
		   'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
		   'Topic :: Software Development'],
      license='BSD'
)




class web_engine(object):
    def __init__(self, processors, req_factory=None, err_hdr=None, uerr_hdr=None):
        super(web_engine, self).__init__()
        self.processors = processors
	if req_factory:
	    self.__req_factory = req_factory
	else:
	    self.__req_factory = request
	    pass
	self.err_hdr = err_hdr
	self.uerr_hdr = uerr_hdr
        pass
    
    def process(self, form, env, input_fo, output_fo, error_fo):
        req = self.__req_factory(form, env, input_fo, output_fo, error_fo)
	try:
	    processors = self.processors
	    for p in processors:
		req = p.ingress(req)
		pass
	    for p in reversed(processors):
		req = p.outgress(req)
		pass
	    pass
	except proc_error, e:
	    self.error(e, req)
	except:
	    self.unknow_error(req)
	    pass
	pass
    
    def error(self, e, req):
	if self.err_hdr:
	    self.err_hdr(e, req)
	    return
	import sys, traceback
	_type, value, tb = sys.exc_info()
	req.output_fo.write('Status: ' + str(e.code) + ' ' + e.msg + '\n')
	req.output_fo.write('Content-Type: text/plain\n\n')
	req.output_fo.write(str(e.code) + ' ' + e.msg +'\n\n')
	traceback.print_tb(tb, file=req.output_fo)
	pass
    
    def unknow_error(self, req):
	if self.uerr_hdr:
	    self.uerr_hdr(req)
	    return
	import sys, traceback
	_type, value, tb = sys.exc_info()
	req.output_fo.write('Status: 500 Internal Server Error\n')
	req.output_fo.write('Content-Type: text/plain\n\n')
	req.output_fo.write(str(req.__dict__))
	req.output_fo.write('\n\n')
	req.output_fo.write(_type.__name__ + ': ' + str(value) + '\n')
	traceback.print_tb(tb, file=req.output_fo)
	pass
    pass


class request(object):
    def __init__(self, form, env, input_fo, output_fo, error_fo):
        super(request, self).__init__()
        self.form = form
        self.env = env
        self.input_fo = input_fo
	self.output_fo = output_fo
	self.error_fo = error_fo
        pass
    
    def created(self, location):
	self.status = 201
	self.location = location
	pass
    
    def no_content(self):
	self.status = 204
	pass
    
    def redirect(self, location):
	self.status = 302
	self.location = location
	pass
    
    def precondition_failed(self):
	self.status = 412
	pass
    
    def method_not_allowed(self):
	self.status = 405
	pass
    
    def not_found(self):
	self.status = 404
	pass
    
    def forbidden(self):
	self.status = 403
	pass
    
    def unauthorized(self):
	self.status = 401
	pass
    
    def bad_request(self):
	self.status = 400
	pass
    pass


class proc_error(Exception):
    def __init__(self, code=404, msg='not found'):
	self.code = code
	self.msg = msg
	pass
    pass


class processor(object):
    def __init__(self):
        super(processor, self).__init__()
        pass
    
    def ingress(self, req):
        return req
    
    def outgress(self, req):
        return req
    pass


class status_proc(processor):
    msg = { 404: '404 not found' }
    def outgress(self, req):
	if hasattr(req, 'status'):
	    req.output_fo.write('Status: ' + msg[req.status] + '\n')
	    pass
	return req
    pass


class header_proc(processor):
    def ingress(self, req):
	req.env['HEADERS_OUT'] = {}
	return req
	pass
    
    def outgress(self, req):
	if req.env['HEADERS_OUT']:
	    headers = req.env['HEADERS_OUT']
	    for k, v in headers.iteritems():
		req.output_fo.write(k + ': ' + v + '\n')
		pass
	    pass
	return req
    pass


class mime_proc(processor):
    def outgress(self, req):
	if hasattr(req, 'mime_type'):
	    mime = req.mime_type
	else:
	    mime = 'text/html'
	    pass
	if mime:
	    req.output_fo.write('Content-Type: ' + mime + '\n\n')
	    pass
	return req;
    pass


class path_to_mod_proc(processor):
    def outgress(self, req):
	from tools import partial_import
	import string
	env = req.env
	try:
	    path = env['PATH_INFO']
	except KeyError:
	    path = '/'
	    pass
	
	pparts = path.split('/')
	pparts = ['ez_controls'] + [part for part in pparts if part]
	
	mod, pparts, i = partial_import(pparts)
	
	if not hasattr(mod, 'process'):
	    raise proc_error(404, 'can not load module "' + string.join(pparts, '.') + '", stop at "' + string.join(pparts[:i], '.') + '"')
	
	saved_path = env['PATH_INFO']
	env['PATH_INFO'] = '/' + string.join(pparts[i:], '/')
	try:
	    mod.process(req)
	finally:
	    env['PATH_INFO'] = saved_path
	    pass
	return req
    pass

